import base64
import os
import unittest

from tonclient.test.test_abi import SAMPLES_DIR
from tonclient.test.helpers import send_grams, async_core_client
from tonclient.types import Abi, DeploySet, CallSet, Signer


class TestTonProcessingAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        self.events_abi = Abi.from_json_path(
            path=os.path.join(SAMPLES_DIR, 'Events.abi.json'))
        with open(os.path.join(SAMPLES_DIR, 'Events.tvc'), 'rb') as fp:
            self.events_tvc = base64.b64encode(fp.read()).decode()
        self.deploy_set = DeploySet(tvc=self.events_tvc)

    def test_process_message(self):
        # Prepare data for deployment message
        keypair = async_core_client.crypto.generate_random_sign_keys()
        signer = Signer.from_keypair(keypair=keypair)
        call_set = CallSet(
            function_name='constructor', header={'pubkey': keypair.public})
        # Encode deployment message
        encoded = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer,
            deploy_set=self.deploy_set, call_set=call_set)

        # Send grams
        send_grams(address=encoded['address'])

        # Deploy account
        result = async_core_client.processing.process_message(
            abi=self.events_abi, signer=signer, deploy_set=self.deploy_set,
            call_set=call_set, send_events=False)

        self.assertEqual([], result['out_messages'])
        self.assertEqual(
            {'out_messages': [], 'output': None}, result['decoded'])

    def test_process_message_with_events(self):
        # Prepare data for deployment message
        keypair = async_core_client.crypto.generate_random_sign_keys()
        signer = Signer.from_keypair(keypair=keypair)
        call_set = CallSet(
            function_name='constructor', header={'pubkey': keypair.public})
        # Encode deployment message
        encoded = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer,
            deploy_set=self.deploy_set, call_set=call_set)

        # Send grams
        send_grams(address=encoded['address'])

        # Deploy account
        generator = async_core_client.processing.process_message(
            abi=self.events_abi, signer=signer, deploy_set=self.deploy_set,
            call_set=call_set, send_events=True)
        events = []
        for event in generator:
            events.append(event)

        result = events[-1]['response_data']
        self.assertEqual([], result['out_messages'])
        self.assertEqual(
            {'out_messages': [], 'output': None}, result['decoded'])

    def test_wait_for_transaction(self):
        # Create deploy message
        keypair = async_core_client.crypto.generate_random_sign_keys()
        signer = Signer.from_keypair(keypair=keypair)
        call_set = CallSet(
            function_name='constructor', header={'pubkey': keypair.public})
        encoded = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer,
            deploy_set=self.deploy_set, call_set=call_set)

        # Send grams
        send_grams(address=encoded['address'])

        # Send message
        shard_block_id = async_core_client.processing.send_message(
            message=encoded['message'], send_events=False, abi=self.events_abi)

        # Wait for transaction
        result = async_core_client.processing.wait_for_transaction(
            message=encoded['message'], shard_block_id=shard_block_id,
            send_events=False, abi=self.events_abi)
        self.assertEqual([], result['out_messages'])
        self.assertEqual(
            {'out_messages': [], 'output': None}, result['decoded'])

    def test_wait_for_transaction_with_events(self):
        # Create deploy message
        keypair = async_core_client.crypto.generate_random_sign_keys()
        signer = Signer.from_keypair(keypair=keypair)
        call_set = CallSet(
            function_name='constructor', header={'pubkey': keypair.public})
        encoded = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer,
            deploy_set=self.deploy_set, call_set=call_set)

        # Send grams
        send_grams(address=encoded['address'])

        # Send message
        generator = async_core_client.processing.send_message(
            message=encoded['message'], send_events=True, abi=self.events_abi)
        events = []
        for event in generator:
            events.append(event)
        shard_block_id = events[-1]['response_data']

        # Wait for transaction
        generator = async_core_client.processing.wait_for_transaction(
            message=encoded['message'], shard_block_id=shard_block_id,
            send_events=True, abi=self.events_abi)
        events.clear()
        for event in generator:
            events.append(event)

        result = events[-1]['response_data']
        self.assertEqual([], result['out_messages'])
        self.assertEqual(
            {'out_messages': [], 'output': None}, result['decoded'])



import asyncio
from datetime import datetime

import aiounittest

from tonclient.client import TonClient, DEVNET_BASE_URL
from tonclient.net import TonQLQuery


class TestTonClientAsync(aiounittest.AsyncTestCase):
    def setUp(self) -> None:
        self.client = TonClient(
            network={'server_address': DEVNET_BASE_URL}, is_async=True)

    async def test_gathering(self):
        subscribe_results, mnemonics = await asyncio.gather(
            self._coro_subscription(), self._coro_mnemonics())
        self.assertGreater(len(subscribe_results), 0)
        self.assertEqual(10, len(mnemonics))

    async def _coro_subscription(self):
        now = int(datetime.now().timestamp())
        query = TonQLQuery(collection='messages') \
            .set_filter(created_at__gt=now).set_result('created_at')
        generator = self.client.net.subscribe_collection(query=query)
        handle = None
        results = []
        async for response in generator:
            results.append(response)
            if response['response_data'].get('handle'):
                handle = response['response_data']['handle']

            if (int(datetime.now().timestamp()) > now + 5 or
                    response['response_type'] > 100) and handle:
                await self.client.net.unsubscribe(handle=handle)
                handle = None
        return results

    async def _coro_mnemonics(self):
        mnemonics = []
        while len(mnemonics) < 10:
            mnemonics.append(
                await self.client.crypto.mnemonic_from_random())
            await asyncio.sleep(1)
        return mnemonics



import base64

import unittest

from tonclient.errors import TonException
from tonclient.test.helpers import async_core_client, sync_core_client
from tonclient.types import KeyPair, MnemonicDictionary


class TestTonCryptoAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        self.mnemonic = 'abuse boss fly battle rubber wasp afraid hamster guide essence vibrant tattoo'
        self.master_xprv = 'xprv9s21ZrQH143K25JhKqEwvJW7QAiVvkmi4WRenBZanA6kxHKtKAQQKwZG65kCyW5jWJ8NY9e3GkRoistUjjcpHNsGBUv94istDPXvqGNuWpC'

    def test_sha256(self):
        data = base64.b64encode('TON is our future'.encode())
        result = async_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

        data = base64.b64encode(
            bytes.fromhex('544f4e206973206f757220667574757265'))
        result = async_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

        data = 'VE9OIGlzIG91ciBmdXR1cmU='
        result = async_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

    def test_sha512(self):
        data = base64.b64encode('TON is our future'.encode())
        result = async_core_client.crypto.sha512(data=data)
        self.assertEqual(
            '4c52dd4cefc68319bac5e97c1f0d18ae8194fb0dd8d9e090ba8376834a0756175a9a736d1e69cb1a58d25c3d554b02a2b8ed9c3ae5cbeeccc3277746a363a434',
            result)

    def test_hdkey_xprv_from_mnemonic(self):
        xprv = async_core_client.crypto.hdkey_xprv_from_mnemonic(
            phrase=self.mnemonic)
        self.assertEqual(self.master_xprv, xprv)

        with self.assertRaises(TonException):
            async_core_client.crypto.hdkey_xprv_from_mnemonic(phrase=0)

    def test_hdkey_secret_from_xprv(self):
        secret = async_core_client.crypto.hdkey_secret_from_xprv(
            xprv=self.master_xprv)
        self.assertEqual(
            '0c91e53128fa4d67589d63a6c44049c1068ec28a63069a55ca3de30c57f8b365',
            secret)

        with self.assertRaises(TonException):
            async_core_client.crypto.hdkey_secret_from_xprv(xprv='')

    def test_hdkey_public_from_xprv(self):
        public = async_core_client.crypto.hdkey_public_from_xprv(
            xprv=self.master_xprv)
        self.assertEqual(
            '02a8eb63085f73c33fa31b4d1134259406347284f8dab6fc68f4bf8c96f6c39b75',
            public)

    def test_hdkey_derive_from_xprv(self):
        child_xprv = async_core_client.crypto.hdkey_derive_from_xprv(
            xprv=self.master_xprv, child_index=0, hardened=False)
        self.assertEqual(
            'xprv9uZwtSeoKf1swgAkVVCEUmC2at6t7MCJoHnBbn1MWJZyxQ4cySkVXPyNh7zjf9VjsP4vEHDDD2a6R35cHubg4WpzXRzniYiy8aJh1gNnBKv',
            child_xprv)

        with self.assertRaises(TonException):
            async_core_client.crypto.hdkey_derive_from_xprv(
                xprv=self.master_xprv, child_index=-1, hardened=False)

    def test_hdkey_derive_from_xprv_path(self):
        xprv = async_core_client.crypto.hdkey_derive_from_xprv_path(
            xprv=self.master_xprv, path="m/44'/60'/0'/0'")
        self.assertEqual(
            'xprvA1KNMo63UcGjmDF1bX39Cw2BXGUwrwMjeD5qvQ3tA3qS3mZQkGtpf4DHq8FDLKAvAjXsYGLHDP2dVzLu9ycta8PXLuSYib2T3vzLf3brVgZ',
            xprv)

        with self.assertRaises(TonException):
            async_core_client.crypto.hdkey_derive_from_xprv_path(
                xprv=self.master_xprv, path='m/')

    def test_convert_public_key_to_ton_safe_format(self):
        safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            public_key='06117f59ade83e097e0fb33e5d29e8735bda82b3bf78a015542aaa853bb69600')
        self.assertEqual(
            'PuYGEX9Zreg-CX4Psz5dKehzW9qCs794oBVUKqqFO7aWAOTD', safe)

        with self.assertRaises(TonException):
            async_core_client.crypto.convert_public_key_to_ton_safe_format(
                public_key=None)

    def test_generate_random_sign_keys(self):
        keypair = async_core_client.crypto.generate_random_sign_keys()
        self.assertEqual(64, len(keypair.public))
        self.assertEqual(64, len(keypair.secret))
        self.assertNotEqual(keypair.secret, keypair.public)

    def test_sign_and_verify(self):
        unsigned = base64.b64encode('Test Message'.encode())
        keypair = KeyPair(
            public='1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e',
            secret='56b6a77093d6fdf14e593f36275d872d75de5b341942376b2a08759f3cbae78f')

        # Sign message
        signed = async_core_client.crypto.sign(unsigned=unsigned, keys=keypair)
        self.assertEqual(
            '+wz+QO6l1slgZS5s65BNqKcu4vz24FCJz4NSAxef9lu0jFfs8x3PzSZRC+pn5k8+aJi3xYMA3BQzglQmjK3hA1Rlc3QgTWVzc2FnZQ==',
            signed['signed'])
        self.assertEqual(
            'fb0cfe40eea5d6c960652e6ceb904da8a72ee2fcf6e05089cf835203179ff65bb48c57ecf31dcfcd26510bea67e64f3e6898b7c58300dc14338254268cade103',
            signed['signature'])

        # Verify signature
        verified = async_core_client.crypto.verify_signature(
            signed=signed['signed'], public=keypair.public)
        self.assertEqual(unsigned.decode(), verified)
        self.assertEqual(
            base64.b64decode(unsigned), base64.b64decode(verified.encode()))

        with self.assertRaises(TonException):
            async_core_client.crypto.sign(
                unsigned=unsigned, keys=KeyPair(public='1', secret='2'))

        with self.assertRaises(TonException):
            async_core_client.crypto.verify_signature(
                signed='simple', public=keypair.public)

    def test_modular_power(self):
        result = async_core_client.crypto.modular_power(
            base='0123456789ABCDEF', exponent='0123', modulus='01234567')
        self.assertEqual('63bfdf', result)

        with self.assertRaises(TonException):
            async_core_client.crypto.modular_power(
                base='1', exponent='0123', modulus='0.2')

    def test_factorize(self):
        factors = async_core_client.crypto.factorize(
            composite='17ED48941A08F981')
        self.assertIsInstance(factors, list)
        self.assertEqual('494C553B', factors[0])
        self.assertEqual('53911073', factors[1])

        with self.assertRaises(TonException):
            async_core_client.crypto.factorize(composite='a3')

    def test_ton_crc16(self):
        data = base64.b64encode(bytes.fromhex('0123456789abcdef'))
        crc = async_core_client.crypto.ton_crc16(data=data)
        self.assertEqual(43349, crc)

        with self.assertRaises(TonException):
            async_core_client.crypto.ton_crc16(data='0==')

    def test_generate_random_bytes(self):
        bytes_b64 = async_core_client.crypto.generate_random_bytes(length=32)
        self.assertEqual(44, len(bytes_b64))
        bts = base64.b64decode(bytes_b64.encode())
        self.assertEqual(32, len(bts))

        with self.assertRaises(TonException):
            async_core_client.crypto.generate_random_bytes(length='1')

    def test_mnemonic_words(self):
        words = async_core_client.crypto.mnemonic_words()
        self.assertEqual(2048, len(words.split(' ')))

        with self.assertRaises(TonException):
            async_core_client.crypto.mnemonic_words(dictionary=100)

    def test_mnemonic_from_random(self):
        for d in range(1, 8):
            for count in [12, 15, 18, 21, 24]:
                mnemonic = async_core_client.crypto.mnemonic_from_random(
                    word_count=count, dictionary=d)
                self.assertEqual(count, len(mnemonic.split(' ')))

        with self.assertRaises(TonException):
            async_core_client.crypto.mnemonic_from_random(word_count=0)

    def test_mnemonic_from_entropy(self):
        mnemonic = async_core_client.crypto.mnemonic_from_entropy(
            entropy='00112233445566778899AABBCCDDEEFF')
        self.assertEqual(
            'abandon math mimic master filter design carbon crystal rookie group knife young',
            mnemonic)

        with self.assertRaises(TonException):
            async_core_client.crypto.mnemonic_from_entropy(entropy='01')

    def test_mnemonic_verify(self):
        for d in range(1, 8):
            for count in [12, 15, 18, 21, 24]:
                mnemonic = async_core_client.crypto.mnemonic_from_random(
                    word_count=count, dictionary=d)
                valid = async_core_client.crypto.mnemonic_verify(
                    phrase=mnemonic, word_count=count, dictionary=d)
                self.assertEqual(True, valid)

        valid = async_core_client.crypto.mnemonic_verify(phrase='one')
        self.assertEqual(False, valid)

    def test_mnemonic_derive_sign_keys(self):
        phrase = 'unit follow zone decline glare flower crisp vocal adapt magic much mesh cherry teach mechanic rain float vicious solution assume hedgehog rail sort chuckle'
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            phrase=phrase, dictionary=MnemonicDictionary.TON, word_count=24)
        public_safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            public_key=keypair.public)
        self.assertEqual(
            'PuYTvCuf__YXhp-4jv3TXTHL0iK65ImwxG0RGrYc1sP3H4KS', public_safe)

        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            phrase=phrase, path='m', dictionary=MnemonicDictionary.TON,
            word_count=24)
        public_safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            public_key=keypair.public)
        self.assertEqual(
            'PubDdJkMyss2qHywFuVP1vzww0TpsLxnRNnbifTCcu-XEgW0', public_safe)

        phrase = 'abandon math mimic master filter design carbon crystal rookie group knife young'
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            phrase=phrase)
        public_safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            public_key=keypair.public)
        self.assertEqual(
            'PuZhw8W5ejPJwKA68RL7sn4_RNmeH4BIU_mEK7em5d4_-cIx', public_safe)

        mnemonic = async_core_client.crypto.mnemonic_from_entropy(
            entropy='2199ebe996f14d9e4e2595113ad1e627')
        keypair = async_core_client.crypto.mnemonic_derive_sign_keys(
            phrase=mnemonic)
        public_safe = async_core_client.crypto.convert_public_key_to_ton_safe_format(
            public_key=keypair.public)
        self.assertEqual(
            'PuZdw_KyXIzo8IksTrERN3_WoAoYTyK7OvM-yaLk711sUIB3', public_safe)

    def test_nacl_sign_keypair_from_secret_key(self):
        keypair = async_core_client.crypto.nacl_sign_keypair_from_secret_key(
            secret='8fb4f2d256e57138fb310b0a6dac5bbc4bee09eb4821223a720e5b8e1f3dd674')
        self.assertEqual(
            'aa5533618573860a7e1bf19f34bd292871710ed5b2eafa0dcdbb33405f2231c6',
            keypair.public)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_sign_keypair_from_secret_key(
                secret='0a')

    def test_nacl_sign(self):
        # Nacl sign data
        unsigned = base64.b64encode('Test Message'.encode())
        secret = '56b6a77093d6fdf14e593f36275d872d75de5b341942376b2a08759f3cbae78f1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e'
        signed = async_core_client.crypto.nacl_sign(
            unsigned=unsigned, secret=secret)
        self.assertEqual(
            '+wz+QO6l1slgZS5s65BNqKcu4vz24FCJz4NSAxef9lu0jFfs8x3PzSZRC+pn5k8+aJi3xYMA3BQzglQmjK3hA1Rlc3QgTWVzc2FnZQ==',
            signed)

        # Nacl sign open
        opened = async_core_client.crypto.nacl_sign_open(
            signed=signed, public='1869b7ef29d58026217e9cf163cbfbd0de889bdf1bf4daebf5433a312f5b8d6e')
        self.assertEqual(unsigned, opened.encode())

        # Nacl sign detached
        signature = async_core_client.crypto.nacl_sign_detached(
            unsigned=unsigned, secret=secret)
        self.assertEqual(
            'fb0cfe40eea5d6c960652e6ceb904da8a72ee2fcf6e05089cf835203179ff65bb48c57ecf31dcfcd26510bea67e64f3e6898b7c58300dc14338254268cade103',
            signature)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_sign(unsigned='0==', secret=secret)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_sign_open(
                signed=signed, public='0x00')

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_sign_detached(
                unsigned=100, secret=secret)

    def test_nacl_box_keypair(self):
        keypair = async_core_client.crypto.nacl_box_keypair()
        self.assertEqual(64, len(keypair.public))
        self.assertEqual(64, len(keypair.secret))
        self.assertNotEqual(keypair.public, keypair.secret)

    def test_nacl_box_keypair_from_secret_key(self):
        keypair = async_core_client.crypto.nacl_box_keypair_from_secret_key(
            secret='e207b5966fb2c5be1b71ed94ea813202706ab84253bdf4dc55232f82a1caf0d4')
        self.assertEqual(
            'a53b003d3ffc1e159355cb37332d67fc235a7feb6381e36c803274074dc3933a',
            keypair.public)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_box_keypair_from_secret_key(
                secret='0x00')

    def test_nacl_box_and_open(self):
        decrypted = base64.b64encode('Test Message'.encode())
        nonce = 'cd7f99924bf422544046e83595dd5803f17536f5c9a11746'
        their_public = 'c4e2d9fe6a6baf8d1812b799856ef2a306291be7a7024837ad33a8530db79c6b'
        secret = 'd9b9dc5033fb416134e5d2107fdbacab5aadb297cb82dbdcd137d663bac59f7f'

        # Create nacl box
        box = async_core_client.crypto.nacl_box(
            decrypted=decrypted, nonce=nonce, their_public=their_public,
            secret=secret)
        self.assertEqual('li4XED4kx/pjQ2qdP0eR2d/K30uN94voNADxwA==', box)

        # Open nacl box
        opened = async_core_client.crypto.nacl_box_open(
            encrypted=box, nonce=nonce, their_public=their_public,
            secret=secret)
        self.assertEqual(decrypted, opened.encode())

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_box(
                decrypted='0x00', nonce=nonce, their_public='', secret=secret)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_box_open(
                encrypted=box, nonce=nonce, their_public=their_public,
                secret='')

    def test_nacl_secret_box_and_open(self):
        decrypted = base64.b64encode('Test Message \' \" {} $=,?'.encode())
        nonce = '2a33564717595ebe53d91a785b9e068aba625c8453a76e45'
        key = '8f68445b4e78c000fe4d6b7fc826879c1e63e3118379219a754ae66327764bd8'

        # Create nacl secret box
        box = async_core_client.crypto.nacl_secret_box(
            decrypted=decrypted, nonce=nonce, key=key)
        self.assertEqual(
            'I6QZteixTdul0K0ldT+/U4QF0t/C1Q8RGyzQ2Hl7886DpW3/DK5ijg==', box)

        # Open nacl secret box
        opened = async_core_client.crypto.nacl_secret_box_open(
            encrypted=box, nonce=nonce, key=key)
        self.assertEqual(decrypted, opened.encode())

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_secret_box(
                decrypted='0x00', nonce=nonce, key=None)

        with self.assertRaises(TonException):
            async_core_client.crypto.nacl_secret_box_open(
                encrypted=box, nonce=nonce, key='')

    def test_scrypt(self):
        password = base64.b64encode('Test Password'.encode())
        salt = base64.b64encode('Test Salt'.encode())
        key = async_core_client.crypto.scrypt(
            password=password, salt=salt, log_n=10, r=8, p=16, dk_len=64)
        self.assertEqual(
            '52e7fcf91356eca55fc5d52f16f5d777e3521f54e3c570c9bbb7df58fc15add73994e5db42be368de7ebed93c9d4f21f9be7cc453358d734b04a057d0ed3626d',
            key)

        with self.assertRaises(TonException):
            async_core_client.crypto.scrypt(
                password=password, salt=salt, log_n=10, r=1, p=1, dk_len=0)


class TestTonCryptoSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def test_sha256(self):
        data = base64.b64encode('TON is our future'.encode())
        result = sync_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

        data = base64.b64encode(
            bytes.fromhex('544f4e206973206f757220667574757265'))
        result = sync_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

        data = 'VE9OIGlzIG91ciBmdXR1cmU='
        result = sync_core_client.crypto.sha256(data=data)
        self.assertEqual(
            '1e7fd5ec201652b5375e5edf3e86d0513394d2c2004dd506415abf0578261951',
            result)

    def test_sha512(self):
        data = base64.b64encode('TON is our future'.encode())
        result = sync_core_client.crypto.sha512(data=data)
        self.assertEqual(
            '4c52dd4cefc68319bac5e97c1f0d18ae8194fb0dd8d9e090ba8376834a0756175a9a736d1e69cb1a58d25c3d554b02a2b8ed9c3ae5cbeeccc3277746a363a434',
            result)

    def test_scrypt(self):
        password = base64.b64encode('Test Password'.encode())
        salt = base64.b64encode('Test Salt'.encode())
        key = sync_core_client.crypto.scrypt(
            password=password, salt=salt, log_n=10, r=8, p=16, dk_len=64)
        self.assertEqual(
            '52e7fcf91356eca55fc5d52f16f5d777e3521f54e3c570c9bbb7df58fc15add73994e5db42be368de7ebed93c9d4f21f9be7cc453358d734b04a057d0ed3626d',
            key)

        with self.assertRaises(TonException):
            sync_core_client.crypto.scrypt(
                password=password, salt=salt, log_n=10, r=1, p=1, dk_len=0)



from typing import Dict, Union, Any

from tonclient.bindings.lib import tc_create_context, tc_destroy_context, \
    tc_read_string, tc_destroy_string
from tonclient.bindings.types import TCClientContext, TCSyncResponseData
from tonclient.boc import TonBoc
from tonclient.decorators import Response
from tonclient.errors import TonException
from tonclient.module import TonModule
from tonclient.crypto import TonCrypto
from tonclient.net import TonNet
from tonclient.abi import TonAbi
from tonclient.processing import TonProcessing
from tonclient.tvm import TonTvm
from tonclient.utils import TonUtils

DEVNET_BASE_URL = 'net.ton.dev'
MAINNET_BASE_URL = 'main.ton.dev'

CLIENT_DEFAULT_SETUP = {
    'network': {
        'server_address': 'http://localhost',
        'message_retries_count': 5,
        'message_processing_timeout': 40000,
        'wait_for_timeout': 40000,
        'out_of_sync_threshold': 15000,
        'access_key': ''
    },
    'crypto': {
        'fish_param': ''
    },
    'abi': {
        'message_expiration_timeout': 40000,
        'message_expiration_timeout_grow_factor': 1.5
    }
}


class TonClientBase(TonModule):
    @Response.version
    def version(self) -> str:
        """ Returns Core Library version """
        return self.request(method='client.version')

    @Response.get_api_reference
    def get_api_reference(self) -> Dict:
        """ Returns Core Library API reference """
        return self.request(method='client.get_api_reference')

    @Response.build_info
    def build_info(self) -> Dict[str, Any]:
        return self.request(method='client.build_info')


class TonClient(object):
    """ Main client class to create object of """
    def __init__(
            self, network: Dict[str, Union[str, int]] = None,
            crypto: Dict[str, str] = None,
            abi: Dict[str, Union[int, float]] = None,
            is_core_async: bool = True, is_async: bool = False):
        """
        :param network: Client core network config
        :param crypto: Client core crypto config
        :param abi: Client core abi config
        :param is_core_async: Use sync or async core requests
        :param is_async: Client mode
        """
        super(TonClient, self).__init__()

        self._ctx = self.create_context(config={
            'network': {**CLIENT_DEFAULT_SETUP['network'], **(network or {})},
            'crypto': {**CLIENT_DEFAULT_SETUP['crypto'], **(crypto or {})},
            'abi': {**CLIENT_DEFAULT_SETUP['abi'], **(abi or {})}
        })
        self._is_core_async = is_core_async
        self._is_async = is_async

        self.base = TonClientBase(client=self)
        self.crypto = TonCrypto(client=self)
        self.net = TonNet(client=self)
        self.abi = TonAbi(client=self)
        self.boc = TonBoc(client=self)
        self.processing = TonProcessing(client=self)
        self.utils = TonUtils(client=self)
        self.tvm = TonTvm(client=self)

    @property
    def ctx(self):
        return self._ctx

    @property
    def is_core_async(self):
        return self._is_core_async

    @is_core_async.setter
    def is_core_async(self, value: bool):
        self._is_core_async = value

    @property
    def is_async(self):
        return self._is_async

    @is_async.setter
    def is_async(self, value: bool):
        self._is_async = value

    @property
    def version(self):
        return self.base.version

    @property
    def get_api_reference(self):
        return self.base.get_api_reference

    @property
    def build_info(self):
        return self.base.build_info

    @staticmethod
    def create_context(config: Dict[str, Dict[str, Union[str, int, float]]]):
        response_ptr = tc_create_context(config=config)
        response = TCSyncResponseData(tc_read_string(string=response_ptr))
        is_success, result, error = (
            response.is_success, response.result, response.error)
        tc_destroy_string(response_ptr)

        if not is_success:
            raise TonException(error=error)

        return TCClientContext(result)

    def destroy_context(self):
        tc_destroy_context(ctx=self._ctx)



from datetime import datetime

import unittest

from tonclient.errors import TonException
from tonclient.net import TonQLQuery
from tonclient.test.helpers import async_core_client, sync_core_client


class TestTonNetAsyncCore(unittest.TestCase):
    def test_query_collection(self):
        query = TonQLQuery(collection='blocks_signatures') \
            .set_result('id').set_limit(1)
        result = async_core_client.net.query_collection(query=query)
        self.assertGreater(len(result), 0)

        query = TonQLQuery(collection='accounts').set_result('id', 'balance') \
            .set_limit(5)
        result = async_core_client.net.query_collection(query=query)
        self.assertEqual(5, len(result))

        query = TonQLQuery(collection='messages') \
            .set_filter(created_at__gt=1562342740) \
            .set_result('body created_at').set_order('created_at') \
            .set_limit(10)
        result = async_core_client.net.query_collection(query=query)
        self.assertGreater(result[0]['created_at'], 1562342740)

        with self.assertRaises(TonException):
            query = TonQLQuery(collection='messages')
            async_core_client.net.query_collection(query=query)

    def test_wait_for_collection(self):
        now = int(datetime.now().timestamp())
        query = TonQLQuery(collection='transactions') \
            .set_filter(now__gt=now).set_result('id now')
        result = async_core_client.net.wait_for_collection(query=query)
        self.assertGreater(result['now'], now)

        with self.assertRaises(TonException):
            async_core_client.net.wait_for_collection(query=query, timeout=1)

    def test_subscribe_collection(self):
        # Prepare query
        now = int(datetime.now().timestamp())
        query = TonQLQuery(collection='messages') \
            .set_filter(created_at__gt=now).set_result('created_at')

        # Create generator
        generator = async_core_client.net.subscribe_collection(query=query)
        handle = None
        results = []
        for response in generator:
            results.append(response)
            if response['response_data'].get('handle'):
                handle = response['response_data']['handle']

            if (int(datetime.now().timestamp()) > now + 5 or
                    response['response_type'] > 100) and handle:
                async_core_client.net.unsubscribe(handle=handle)
                handle = None

        self.assertGreater(len(results), 0)


class TestTonNetSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def test_query_collection(self):
        query = TonQLQuery(collection='blocks_signatures') \
            .set_result('id').set_limit(1)
        result = sync_core_client.net.query_collection(query=query)
        self.assertGreater(len(result), 0)

        query = TonQLQuery(collection='accounts').set_result('id', 'balance') \
            .set_limit(5)
        result = sync_core_client.net.query_collection(query=query)
        self.assertEqual(5, len(result))

        query = TonQLQuery(collection='messages') \
            .set_filter(created_at__gt=1562342740) \
            .set_result('body created_at').set_order('created_at') \
            .set_limit(10)
        result = sync_core_client.net.query_collection(query=query)
        self.assertGreater(result[0]['created_at'], 1562342740)

        with self.assertRaises(TonException):
            query = TonQLQuery(collection='messages')
            sync_core_client.net.query_collection(query=query)

    def test_wait_for_collection(self):
        now = int(datetime.now().timestamp())
        query = TonQLQuery(collection='transactions') \
            .set_filter(now__gt=now).set_result('id now')
        result = sync_core_client.net.wait_for_collection(query=query)
        self.assertGreater(result['now'], now)

        with self.assertRaises(TonException):
            sync_core_client.net.wait_for_collection(query=query, timeout=1)



from typing import Dict, Any

from tonclient.decorators import Response
from tonclient.module import TonModule
from tonclient.types import Abi, AccountForExecutor


class TonTvm(TonModule):
    """ Free TON tvm SDK API implementation """
    @Response.run_get
    def run_get(
            self, account: str, function_name: str,
            inputs: Any = None, blockchain_config: str = None,
            block_time: int = None, block_lt: int = None,
            transaction_lt: int = None) -> Any:
        """
        Executes get method and returns data from TVM stack
        :param account: Base64 encoded account BOC
        :param function_name:
        :param inputs:
        :param blockchain_config: Execution options: blockchain config BOC
        :param block_time: Execution options: time that is used as
                transaction time
        :param block_lt: Execution options: block logical time
        :param transaction_lt: Execution options: transaction logical time
        :return:
        """
        execution_options = {
            'blockchain_config': blockchain_config,
            'block_time': block_time,
            'block_lt': block_lt,
            'transaction_lt': transaction_lt
        }
        return self.request(
            method='tvm.run_get', account=account,
            function_name=function_name, input=inputs,
            execution_options=execution_options)

    def run_executor(
            self, message: str, account: AccountForExecutor, abi: Abi = None,
            skip_transaction_check: bool = False,
            blockchain_config: str = None, block_time: int = None,
            block_lt: int = None, transaction_lt: int = None
    ) -> Dict[str, Any]:
        """
        :param message: Base64 encoded message BOC
        :param account: Account to run on executor
        :param abi: Contract ABI for decoding output messages
        :param skip_transaction_check: Skip transaction check flag
        :param blockchain_config: Execution options: blockchain config BOC
        :param block_time: Execution options: time that is used as
                transaction time
        :param block_lt: Execution options: block logical time
        :param transaction_lt: Execution options: transaction logical time
        :return:
        """
        abi = abi.dict if abi else abi
        execution_options = {
            'blockchain_config': blockchain_config,
            'block_time': block_time,
            'block_lt': block_lt,
            'transaction_lt': transaction_lt
        }
        return self.request(
            method='tvm.run_executor', message=message, account=account.dict,
            abi=abi, skip_transaction_check=skip_transaction_check,
            execution_options=execution_options)

    def run_tvm(
            self, message: str, account: str, abi: Abi = None,
            blockchain_config: str = None, block_time: int = None,
            block_lt: int = None, transaction_lt: int = None
    ) -> Dict[str, Any]:
        """
        :param message: Base64 encoded message BOC
        :param account: Base64 encoded account BOC
        :param abi: Contract ABI for decoding output messages
        :param blockchain_config: Execution options: blockchain config BOC
        :param block_time: Execution options: time that is used as
                transaction time
        :param block_lt: Execution options: block logical time
        :param transaction_lt: Execution options: transaction logical time
        :return:
        """
        abi = abi.dict if abi else abi
        execution_options = {
            'blockchain_config': blockchain_config,
            'block_time': block_time,
            'block_lt': block_lt,
            'transaction_lt': transaction_lt
        }
        return self.request(
            method='tvm.run_tvm', message=message, account=account, abi=abi,
            execution_options=execution_options)



from typing import Union, Dict, Any, Generator

from tonclient.decorators import Response
from tonclient.module import TonModule
from tonclient.types import Abi, MessageSource, Signer, DeploySet, CallSet


class TonProcessing(TonModule):
    """ Free TON processing SDK API implementation """
    def process_message(
            self, abi: Abi, signer: Signer, send_events: bool,
            address: str = None, deploy_set: DeploySet = None,
            call_set: CallSet = None, processing_try_index: int = 0
    ) -> Union[Dict[str, Any], Generator]:
        """
        Creates message, sends it to the network and monitors its processing
        :param abi: Contract ABI
        :param signer: Signing parameters
        :param address: Contract address. Must be specified in case of
                non deploy message
        :param deploy_set: Deploy parameters. Must be specified in case of
                deploy message
        :param call_set: Function call parameters. Must be specified in
                non deploy message. In case of deploy message contains
                parameters of constructor
        :param processing_try_index: Processing try index. Used in message
                processing with retries. Encoder uses the provided try index
                to calculate message expiration time. Expiration timeouts will
                grow with every retry
        :param send_events: Flag for requesting events sending
        :return:
        """
        message_source = MessageSource.from_encoding_params(
            abi=abi, signer=signer, address=address, deploy_set=deploy_set,
            call_set=call_set, processing_try_index=processing_try_index)
        return self.request(
            method='processing.process_message', as_iterable=send_events,
            message_encode_params=message_source.dict,
            send_events=send_events)

    @Response.send_message
    def send_message(
            self, message: str, send_events: bool, abi: Abi = None
    ) -> Union[str, Generator]:
        """
        Sends message to the network
        :param message: Message BOC
        :param send_events: Flag for requesting events sending
        :param abi: Optional message ABI. If this parameter is specified and
                the message has the 'expire' header then expiration time will
                be checked against the current time to prevent an unnecessary
                sending.
                The `message already expired` error will be returned in this
                case.
                Note that specifying `abi` for ABI compliant contracts is
                strongly recommended due to choosing proper processing strategy
        :return:
        """
        abi = abi.dict if abi else abi
        return self.request(
            method='processing.send_message', message=message,
            send_events=send_events, abi=abi, as_iterable=send_events)

    def wait_for_transaction(
            self, message: str, shard_block_id: str, send_events: bool,
            abi: Abi = None) -> Union[Dict[str, Any], Generator]:
        """
        Performs monitoring of the network for the result transaction
        :param message: Base64 encoded message BOC
        :param shard_block_id: Dst account shard block id before the message
                had been sent. You must provide the same value as the
                'send_message' has returned
        :param send_events: Flag for requesting events sending
        :param abi: Optional ABI for decoding transaction results. If it is
                specified then the output messages bodies will be decoded
                according to this ABI.
        :return:
        """
        abi = abi.dict if abi else abi
        return self.request(
            method='processing.wait_for_transaction', message=message,
            shard_block_id=shard_block_id, send_events=send_events, abi=abi,
            as_iterable=send_events)



import functools
import inspect
from typing import Callable, Any

from tonclient.types import KeyPair


class Response(object):
    @staticmethod
    def __pretty(function, _callback: Callable[[Any], Any] = None):
        """ Base decorator function """
        @functools.wraps(function)
        def response_or_generator(*args, **kwargs):
            def generator_wrapper():
                for item in result:
                    item['response_data'] = _callback(item['response_data'])
                    yield item

            def response_wrapper():
                return _callback(result)

            async def async_response_wrapper():
                _result = await result
                return _callback(_result)

            async def async_generator_wrapper():
                async for item in result:
                    item['response_data'] = _callback(item['response_data'])
                    yield item

            result = function(*args, **kwargs)
            if inspect.isgenerator(result):
                return generator_wrapper()
            elif inspect.iscoroutine(result):
                return async_response_wrapper()
            elif inspect.isasyncgen(result):
                return async_generator_wrapper()
            return response_wrapper()

        return response_or_generator

    @classmethod
    def version(cls, function):
        def __callback(result):
            return result['version']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def get_api_reference(cls, function):
        def __callback(result):
            return result['api']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def build_info(cls, function):
        def __callback(result):
            return result['build_info']

        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def sha256(cls, function):
        def __callback(result):
            return result['hash']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def sha512(cls, function):
        def __callback(result):
            return result['hash']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def hdkey_xprv_from_mnemonic(cls, function):
        def __callback(result):
            return result['xprv']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def hdkey_secret_from_xprv(cls, function):
        def __callback(result):
            return result['secret']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def hdkey_public_from_xprv(cls, function):
        def __callback(result):
            return result['public']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def hdkey_derive_from_xprv(cls, function):
        def __callback(result):
            return result['xprv']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def hdkey_derive_from_xprv_path(cls, function):
        def __callback(result):
            return result['xprv']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def convert_public_key_to_ton_safe_format(cls, function):
        def __callback(result):
            return result['ton_public_key']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def generate_random_sign_keys(cls, function):
        def __callback(result):
            return KeyPair(**result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def verify_signature(cls, function):
        def __callback(result):
            return result['unsigned']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def modular_power(cls, function):
        def __callback(result):
            return result['modular_power']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def factorize(cls, function):
        def __callback(result):
            return result['factors']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def ton_crc16(cls, function):
        def __callback(result):
            return result['crc']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def generate_random_bytes(cls, function):
        def __callback(result):
            return result['bytes']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def mnemonic_words(cls, function):
        def __callback(result):
            return result['words']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def mnemonic_from_random(cls, function):
        def __callback(result):
            return result['phrase']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def mnemonic_from_entropy(cls, function):
        def __callback(result):
            return result['phrase']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def mnemonic_verify(cls, function):
        def __callback(result):
            return result['valid']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def mnemonic_derive_sign_keys(cls, function):
        def __callback(result):
            return KeyPair(**result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_sign_keypair_from_secret_key(cls, function):
        def __callback(result):
            return KeyPair(**result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_sign(cls, function):
        def __callback(result):
            return result['signed']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_sign_detached(cls, function):
        def __callback(result):
            return result['signature']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_sign_open(cls, function):
        def __callback(result):
            return result['unsigned']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_box_keypair(cls, function):
        def __callback(result):
            return KeyPair(**result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_box_keypair_from_secret_key(cls, function):
        def __callback(result):
            return KeyPair(**result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_box(cls, function):
        def __callback(result):
            return result['encrypted']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_box_open(cls, function):
        def __callback(result):
            return result['decrypted']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_secret_box(cls, function):
        def __callback(result):
            return result['encrypted']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def nacl_secret_box_open(cls, function):
        def __callback(result):
            return result['decrypted']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def scrypt(cls, function):
        def __callback(result):
            return result['key']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def query_collection(cls, function):
        def __callback(result):
            return result['result']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def wait_for_collection(cls, function):
        def __callback(result):
            return result['result']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def subscribe_collection(cls, function):
        def __callback(result):
            return result.get('result', result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def parse_message(cls, function):
        def __callback(result):
            return result['parsed']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def parse_transaction(cls, function):
        def __callback(result):
            return result['parsed']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def parse_account(cls, function):
        def __callback(result):
            return result['parsed']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def parse_block(cls, function):
        def __callback(result):
            return result['parsed']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def get_blockchain_config(cls, function):
        def __callback(result):
            return result['config_boc']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def convert_address(cls, function):
        def __callback(result):
            return result['address']
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def send_message(cls, function):
        def __callback(result):
            return result.get('shard_block_id', result)
        return cls.__pretty(function=function, _callback=__callback)

    @classmethod
    def run_get(cls, function):
        def __callback(result):
            return result['output']
        return cls.__pretty(function=function, _callback=__callback)



import os

from tonclient.client import TonClient, DEVNET_BASE_URL
from tonclient.types import Abi, CallSet, Signer

BASE_DIR = os.path.dirname(__file__)
SAMPLES_DIR = os.path.join(BASE_DIR, 'samples')
GIVER_ADDRESS = '0:653b9a6452c7a982c6dc92b2da9eba832ade1c467699ebb3b43dca6d77b780dd'

async_core_client = TonClient(network={'server_address': DEVNET_BASE_URL})
sync_core_client = TonClient(
    network={'server_address': DEVNET_BASE_URL}, is_core_async=False)


def send_grams(address: str):
    giver_abi = Abi.from_json_path(
        path=os.path.join(SAMPLES_DIR, 'Giver.abi.json'))
    call_set = CallSet(
        function_name='grant',
        inputs={'addr': address})
    async_core_client.processing.process_message(
        abi=giver_abi, signer=Signer.none(), address=GIVER_ADDRESS,
        call_set=call_set, send_events=False)



from typing import Dict, Union, List

from tonclient.decorators import Response
from tonclient.module import TonModule
from tonclient.types import KeyPair, DEFAULT_MNEMONIC_DICTIONARY, \
    DEFAULT_MNEMONIC_WORD_COUNT, DEFAULT_HDKEY_DERIVATION_PATH


class TonCrypto(TonModule):
    """ Free TON crypto SDK API implementation """
    @Response.sha256
    def sha256(self, data: Union[str, bytes]) -> str:
        """
        Calculates SHA256 hash of the specified data.
        :param data: Base64 encoded data for hash calculation
        :return:
        """
        if type(data) is bytes:
            data = data.decode()
        return self.request(method='crypto.sha256', data=data)

    @Response.sha512
    def sha512(self, data: Union[str, bytes]) -> str:
        """
        Calculates SHA512 hash of the specified data.
        :param data: Base64 encoded data for hash calculation
        :return:
        """
        if type(data) is bytes:
            data = data.decode()
        return self.request(method='crypto.sha512', data=data)

    @Response.hdkey_xprv_from_mnemonic
    def hdkey_xprv_from_mnemonic(self, phrase: str) -> str:
        """
        Generate the extended master private key that will be the root for
        all the derived keys.
        :param phrase: String with seed phrase
        :return:
        """
        return self.request(
            method='crypto.hdkey_xprv_from_mnemonic', phrase=phrase)

    @Response.hdkey_secret_from_xprv
    def hdkey_secret_from_xprv(self, xprv: str) -> str:
        """
        Extracts the private key from the serialized extended private key.
        :param xprv: Serialized extended private key
        :return:
        """
        return self.request(
            method='crypto.hdkey_secret_from_xprv', xprv=xprv)

    @Response.hdkey_public_from_xprv
    def hdkey_public_from_xprv(self, xprv: str) -> str:
        """
        Extracts the public key from the serialized extended private key.
        :param xprv: Serialized extended private key
        :return:
        """
        return self.request(
            method='crypto.hdkey_public_from_xprv', xprv=xprv)

    @Response.hdkey_derive_from_xprv
    def hdkey_derive_from_xprv(
            self, xprv: str, child_index: int, hardened: bool) -> str:
        """
        Returns derived extended private key derived from the specified
        extended private key and child index
        :param xprv: Serialized extended private key
        :param child_index: Child index (see BIP-0032)
        :param hardened: Indicates the derivation of hardened/not-hardened
                key (see BIP-0032)
        :return:
        """
        return self.request(
            method='crypto.hdkey_derive_from_xprv', xprv=xprv,
            child_index=child_index, hardened=hardened)

    @Response.hdkey_derive_from_xprv_path
    def hdkey_derive_from_xprv_path(
            self, xprv: str, path: str = DEFAULT_HDKEY_DERIVATION_PATH) -> str:
        """
        Derives the extended private key from the specified key and path.
        :param xprv: Serialized extended private key
        :param path: Derivation path, for instance "m/44'/396'/0'/0/0"
        :return:
        """
        return self.request(
            method='crypto.hdkey_derive_from_xprv_path', xprv=xprv,
            path=path)

    @Response.convert_public_key_to_ton_safe_format
    def convert_public_key_to_ton_safe_format(self, public_key: str) -> str:
        """
        Converts public key to ton safe_format.
        :param public_key: Public key
        :return:
        """
        return self.request(
            method='crypto.convert_public_key_to_ton_safe_format',
            public_key=public_key)

    @Response.generate_random_sign_keys
    def generate_random_sign_keys(self) -> KeyPair:
        """
        Generates random ed25519 key pair.
        :return:
        """
        return self.request(method='crypto.generate_random_sign_keys')

    def sign(
            self, unsigned: Union[str, bytes], keys: KeyPair
    ) -> Dict[str, str]:
        """
        Signs a data using the provided keys.
        :param unsigned: Base64 encoded data to be signed
        :param keys: Sign keys
        :return:
        """
        if type(unsigned) is bytes:
            unsigned = unsigned.decode()
        return self.request(
            method='crypto.sign', unsigned=unsigned, keys=keys.dict)

    @Response.verify_signature
    def verify_signature(self, signed: Union[str, bytes], public: str) -> str:
        """
        Verifies signed data using the provided public key.
        :param signed: Base64 encoded signature to be verified
        :param public: Signer's public key hex
        :return: Base64 encoded unsigned data
        """
        if type(signed) is bytes:
            signed = signed.decode()
        return self.request(
            method='crypto.verify_signature', signed=signed,
            public=public)

    @Response.modular_power
    def modular_power(self, base: str, exponent: str, modulus: str) -> str:
        """
        Performs modular exponentiation for big integers
        (`base`^`exponent` mod `modulus`).
        See [https://en.wikipedia.org/wiki/Modular_exponentiation]
        :param base: 'base' argument of calculation
        :param exponent: 'exponent' argument of calculation
        :param modulus: 'modulus' argument of calculation
        :return:
        """
        return self.request(
            method='crypto.modular_power', base=base, exponent=exponent,
            modulus=modulus)

    @Response.factorize
    def factorize(self, composite: str) -> List[str]:
        """
        Performs prime factorization – decomposition of a composite number
        into a product of smaller prime integers (factors).
        See [https://en.wikipedia.org/wiki/Integer_factorization]
        :param composite: Hexadecimal representation of u64 composite number
        :return: Two factors of composite or empty if composite
                can't be factorized
        """
        return self.request(
            method='crypto.factorize', composite=composite)

    @Response.ton_crc16
    def ton_crc16(self, data: Union[str, bytes]) -> int:
        """
        Calculates CRC16 using TON algorithm.
        :param data: Base64 encoded data for CRC calculation
        :return:
        """
        if type(data) is bytes:
            data = data.decode()
        return self.request(method='crypto.ton_crc16', data=data)

    @Response.generate_random_bytes
    def generate_random_bytes(self, length: int) -> str:
        """
        Generates random byte array of the specified length and returns
        it in base64 format
        :param length: Size of random byte array
        :return: Base64 encoded generated bytes
        """
        return self.request(
            method='crypto.generate_random_bytes', length=length)

    @Response.mnemonic_words
    def mnemonic_words(
            self, dictionary: int = DEFAULT_MNEMONIC_DICTIONARY) -> str:
        """
        Prints the list of words from the specified dictionary.
        :param dictionary: Dictionary identifier
        :return: String of dictionary words
        """
        return self.request(
            method='crypto.mnemonic_words', dictionary=dictionary)

    @Response.mnemonic_from_random
    def mnemonic_from_random(
            self, dictionary: int = DEFAULT_MNEMONIC_DICTIONARY,
            word_count: int = DEFAULT_MNEMONIC_WORD_COUNT) -> str:
        """
        Generates a random mnemonic from the specified dictionary
        and word count.
        :param dictionary: Dictionary identifier
        :param word_count: Mnemonic word count
        :return: String of mnemonic words
        """
        return self.request(
            method='crypto.mnemonic_from_random', dictionary=dictionary,
            word_count=word_count)

    @Response.mnemonic_from_entropy
    def mnemonic_from_entropy(
            self, entropy: str, dictionary: int = DEFAULT_MNEMONIC_DICTIONARY,
            word_count: int = DEFAULT_MNEMONIC_WORD_COUNT) -> str:
        """
        Generates mnemonic from pre-generated entropy.
        :param entropy: Hex encoded entropy bytes
        :param dictionary: Dictionary identifier
        :param word_count: Mnemonic word count
        :return: String of mnemonic words
        """
        return self.request(
            method='crypto.mnemonic_from_entropy', entropy=entropy,
            dictionary=dictionary, word_count=word_count)

    @Response.mnemonic_verify
    def mnemonic_verify(
            self, phrase: str, dictionary: int = DEFAULT_MNEMONIC_DICTIONARY,
            word_count: int = DEFAULT_MNEMONIC_WORD_COUNT) -> bool:
        """
        The phrase supplied will be checked for word length and validated
        according to the checksum specified in BIP0039.
        :param phrase: Mnemonic phrase
        :param dictionary: Dictionary identifier
        :param word_count: Mnemonic word count
        :return:
        """
        return self.request(
            method='crypto.mnemonic_verify', phrase=phrase,
            dictionary=dictionary, word_count=word_count)

    @Response.mnemonic_derive_sign_keys
    def mnemonic_derive_sign_keys(
            self, phrase: str, path: str = DEFAULT_HDKEY_DERIVATION_PATH,
            dictionary: int = DEFAULT_MNEMONIC_DICTIONARY,
            word_count: int = DEFAULT_MNEMONIC_WORD_COUNT) -> KeyPair:
        """
        Validates the seed phrase, generates master key and then derives
        the key pair from the master key and the specified path.
        :param phrase: Mnemonic phrase
        :param path: Derivation path, for instance "m/44'/396'/0'/0/0"
        :param dictionary: Dictionary identifier
        :param word_count: Mnemonic word count
        :return:
        """
        return self.request(
            method='crypto.mnemonic_derive_sign_keys', phrase=phrase,
            path=path, dictionary=dictionary, word_count=word_count)

    @Response.nacl_sign_keypair_from_secret_key
    def nacl_sign_keypair_from_secret_key(self, secret: str) -> KeyPair:
        """
        Generates a key pair for signing from the secret key.
        :param secret: Secret key
        :return:
        """
        return self.request(
            method='crypto.nacl_sign_keypair_from_secret_key',
            secret=secret)

    @Response.nacl_sign
    def nacl_sign(self, unsigned: Union[str, bytes], secret: str) -> str:
        """
        Signs data using the signer's secret key.
        :param unsigned: Base64 encoded data to be signed
        :param secret: Signer's secret key
        :return: Base64 encoded signed data
        """
        if type(unsigned) is bytes:
            unsigned = unsigned.decode()
        return self.request(
            method='crypto.nacl_sign', unsigned=unsigned, secret=secret)

    @Response.nacl_sign_detached
    def nacl_sign_detached(
            self, unsigned: Union[str, bytes], secret: str) -> str:
        """
        :param unsigned: Base64 encoded data to be signed
        :param secret: Signer's secret key
        :return: Hex encoded sign
        """
        if type(unsigned) is bytes:
            unsigned = unsigned.decode()
        return self.request(
            method='crypto.nacl_sign_detached', unsigned=unsigned,
            secret=secret)

    @Response.nacl_sign_open
    def nacl_sign_open(self, signed: Union[str, bytes], public: str) -> str:
        """
        :param signed: Base64 encoded signed data to be unsigned
        :param public: Signer's public key
        :return: Base64 encoded unsigned data
        """
        if type(signed) is bytes:
            signed = signed.decode()
        return self.request(
            method='crypto.nacl_sign_open', signed=signed,
            public=public)

    @Response.nacl_box_keypair
    def nacl_box_keypair(self) -> KeyPair:
        return self.request(method='crypto.nacl_box_keypair')

    @Response.nacl_box_keypair_from_secret_key
    def nacl_box_keypair_from_secret_key(self, secret: str) -> KeyPair:
        """
        Generates key pair from a secret key
        :param secret: Hex encoded secret key
        :return:
        """
        return self.request(
            method='crypto.nacl_box_keypair_from_secret_key',
            secret=secret)

    @Response.nacl_box
    def nacl_box(
            self, decrypted: Union[str, bytes], nonce: str, their_public: str,
            secret: str) -> str:
        """
        Public key authenticated encryption
        :param decrypted: Base64 encoded data to be encrypted
        :param nonce:
        :param their_public:
        :param secret:
        :return: Base64 encoded encrypted data
        """
        if type(decrypted) is bytes:
            decrypted = decrypted.decode()
        return self.request(
            method='crypto.nacl_box', decrypted=decrypted, nonce=nonce,
            their_public=their_public, secret=secret)

    @Response.nacl_box_open
    def nacl_box_open(
            self, encrypted: Union[str, bytes], nonce: str, their_public: str,
            secret: str) -> str:
        """
        Decrypt and verify the cipher text using the receiver's secret key
        and sender's public
        :param encrypted: Base64 encoded encrypted data to be decrypted
        :param nonce:
        :param their_public:
        :param secret:
        :return: Base64 encoded decrypted data
        """
        if type(encrypted) is bytes:
            encrypted = encrypted.decode()
        return self.request(
            method='crypto.nacl_box_open', encrypted=encrypted,
            nonce=nonce, their_public=their_public, secret=secret)

    @Response.nacl_secret_box
    def nacl_secret_box(
            self, decrypted: Union[str, bytes], nonce: str, key: str) -> str:
        """
        Encrypt and authenticate message using nonce and secret key
        :param decrypted: Base64 encoded data to be encrypted
        :param nonce:
        :param key:
        :return: Base64 encoded encrypted data
        """
        if type(decrypted) is bytes:
            decrypted = decrypted.decode()
        return self.request(
            method='crypto.nacl_secret_box', decrypted=decrypted,
            nonce=nonce, key=key)

    @Response.nacl_secret_box_open
    def nacl_secret_box_open(
            self, encrypted: Union[str, bytes], nonce: str, key: str) -> str:
        """
        Decrypts and verifies cipher text using nonce and secret key
        :param encrypted: Base64 encoded encrypted data to be decrypted
        :param nonce:
        :param key:
        :return: Base64 encoded decrypted data
        """
        if type(encrypted) is bytes:
            encrypted = encrypted.decode()
        return self.request(
            method='crypto.nacl_secret_box_open', encrypted=encrypted,
            nonce=nonce, key=key)

    @Response.scrypt
    def scrypt(
            self, password: Union[str, bytes], salt: Union[str, bytes],
            log_n: int, r: int, p: int, dk_len: int) -> str:
        """
        Derives key from 'password' and 'key' using 'scrypt' algorithm.
        See [https://en.wikipedia.org/wiki/Scrypt].
        :param password: Base64 encoded password
        :param salt: A base64 encoded salt bytes that modifies the hash to
                protect against Rainbow table attacks
        :param log_n: CPU/memory cost parameter
        :param r: The block size parameter, which fine-tunes sequential
                memory read size and performance (8 is commonly used)
        :param p: Parallelization parameter
        :param dk_len: Intended output length in octets of the derived key
        :return: Hex encoded derived key
        """
        if type(password) is bytes:
            password = password.decode()
        if type(salt) is bytes:
            salt = salt.decode()
        return self.request(
            method='crypto.scrypt', password=password, salt=salt,
            log_n=log_n, r=r, p=p, dk_len=dk_len)



from typing import Dict, Any

from tonclient.module import TonModule
from tonclient.types import Abi, Signer, DeploySet, CallSet, StateInitSource


class TonAbi(TonModule):
    """ Free TON abi SDK API implementation """
    def decode_message(self, abi: Abi, message: str) -> Dict[str, Any]:
        """
        Decodes message body using provided message BOC and ABI
        :param abi: Contract ABI
        :param message: Message BOC
        :return:
        """
        return self.request(
            method='abi.decode_message', abi=abi.dict, message=message)

    def decode_message_body(
            self, abi: Abi, body: str, is_internal: bool = False
    ) -> Dict[str, Any]:
        """
        Decodes message body using provided body BOC and ABI
        :param abi: Contract ABI used to decode
        :param body: Base64 encoded message body BOC
        :param is_internal: True if the body belongs to the internal message
        :return:
        """
        return self.request(
            method='abi.decode_message_body', abi=abi.dict, body=body,
            is_internal=is_internal)

    def encode_account(
            self, state_init: StateInitSource, balance: int = None,
            last_trans_lt: int = None, last_paid: int = None
    ) -> Dict[str, str]:
        """
        Creates account state BOC
        :param state_init: Source of the account state init
        :param balance: Initial balance
        :param last_trans_lt: Initial value for the 'last_trans_lt'
        :param last_paid: Initial value for the 'last_paid'
        :return:
        """
        return self.request(
            method='abi.encode_account', state_init=state_init.dict,
            balance=balance, last_trans_lt=last_trans_lt, last_paid=last_paid)

    def encode_message(
            self, abi: Abi, signer: Signer, address: str = None,
            deploy_set: DeploySet = None, call_set: CallSet = None,
            processing_try_index: int = 0) -> Dict[str, str]:
        """
        Encodes an ABI-compatible message
        :param abi: Contract ABI
        :param address: Contract address. Must be specified in case of
                non deploy message
        :param deploy_set: Deploy parameters. Must be specified in case of
                deploy message
        :param call_set: Function call parameters. Must be specified in
                non deploy message. In case of deploy message contains
                parameters of constructor
        :param signer: Signing parameters
        :param processing_try_index: Processing try index. Used in message
                processing with retries. Encoder uses the provided try index
                to calculate message expiration time. Expiration timeouts will
                grow with every retry
        :return:
        """
        deploy_set = deploy_set.dict if deploy_set else deploy_set
        call_set = call_set.dict if call_set else call_set
        return self.request(
            method='abi.encode_message', abi=abi.dict,
            address=address, deploy_set=deploy_set, call_set=call_set,
            signer=signer.dict, processing_try_index=processing_try_index)

    def encode_message_body(
            self, abi: Abi, call_set: CallSet, signer: Signer,
            is_internal: bool, processing_try_index: int = 0
    ) -> Dict[str, str]:
        """
        Encodes message body according to ABI function call
        :param abi: Contract ABI
        :param call_set: Function call parameters. Must be specified in
                non deploy message. In case of deploy message contains
                parameters of constructor
        :param signer: Signing parameters
        :param is_internal: True if internal message body must be encoded
        :param processing_try_index: Processing try index. Used in message
                processing with retries. Encoder uses the provided try index
                to calculate message expiration time. Expiration timeouts will
                grow with every retry
        :return:
        """
        return self.request(
            method='abi.encode_message_body', abi=abi.dict,
            call_set=call_set.dict, signer=signer.dict,
            is_internal=is_internal, processing_try_index=processing_try_index)

    def attach_signature(
            self, abi: Abi, public_key: str, message: str, signature: str
    ) -> Dict[str, str]:
        """
        Combines hex-encoded signature with base64-encoded unsigned message
        :param abi: Contract ABI
        :param public_key: Hex encoded public key
        :param message: Base64 encoded unsigned message BOC
        :param signature: Hex encoded signature
        :return:
        """
        return self.request(
            method='abi.attach_signature', abi=abi.dict,
            public_key=public_key, message=message, signature=signature)

    def attach_signature_to_message_body(
            self, abi: Abi, public_key: str, message: str, signature: str
    ) -> Dict[str, str]:
        """
        :param abi: Contract ABI
        :param public_key: Hex encoded public key
        :param message: Base64 encoded unsigned message BOC
        :param signature: Hex encoded signature
        :return:
        """
        return self.request(
            method='abi.attach_signature_to_message_body', abi=abi.dict,
            public_key=public_key, message=message, signature=signature)



from typing import Union, Dict

from tonclient.decorators import Response
from tonclient.module import TonModule


class TonUtils(TonModule):
    """ Free TON utils SDK API implementation """
    @Response.convert_address
    def convert_address(
            self, address: str, output_format: Union[str, Dict[str, bool]]
    ) -> str:
        """
        :param address: Account address in any format
        :param output_format: Format to convert to.
                'types.AddressStringFormat' class may help
        :return:
        """
        return self.request(
            method='utils.convert_address', address=address,
            output_format=output_format)



import asyncio
import logging
import os

import json
import time
from typing import Any, Dict, Union, Generator

from tonclient.bindings.lib import tc_request, tc_request_sync, \
    tc_read_string, tc_destroy_string
from tonclient.bindings.types import TCStringData, TCResponseHandler, \
    TCResponseType, TCSyncResponseData
from tonclient.errors import TonException


class TonModule(object):
    """
    Base TON Module class.
    All modules, such as 'crypto', 'contracts', etc. should be inherited
    from this class.
    """
    _async_response_map = {}

    def __init__(self, client):
        self._client = client

    def request(
            self, method: str, as_iterable: bool = False,
            params_or_str: Union[str, Dict[str, Any]] = None, **kwargs) -> Any:
        """ Perform core request """
        # Prepare request params
        request_params = self._prepare_params(params_or_str, **kwargs)
        kwargs = {
            'method': method,
            'request_params': request_params
        }

        # Make sync or async core/client request
        if self._client.is_core_async:
            kwargs.update({'as_iterable': as_iterable})
            return self._async_core_request(**kwargs)
        return self._sync_core_request(**kwargs)

    def _sync_core_request(self, method: str, request_params: str) -> Any:
        """ Perform core synchronous request """
        # Make sync request, get response pointer and parse it
        response_ptr = tc_request_sync(
            ctx=self._client.ctx, method=method, params_json=request_params)
        response = TCSyncResponseData(tc_read_string(string=response_ptr))

        # Copy response data and destroy response pointer
        is_success, result, error = (
            response.is_success, response.result, response.error)
        tc_destroy_string(string=response_ptr)

        if not is_success:
            raise TonException(error=error)

        return result

    def _async_core_request(
            self, method: str, request_params: str, as_iterable: bool
    ) -> Union[Generator, Any]:
        """ Perform core asynchronous request """
        # Perform async core request
        request_id = self._generate_request_id()
        self._async_response_map[request_id] = []
        tc_request(
            ctx=self._client.ctx, method=method, request_id=request_id,
            params_json=request_params,
            response_handler=self._async_response_handler)

        if as_iterable:
            if self._client.is_async:
                return self._async_response_generator_coro(
                    request_id=request_id)
            return self._async_response_generator(request_id=request_id)

        if self._client.is_async:
            return self._async_response_resolver_coro(request_id=request_id)
        return self._async_response_resolver(request_id=request_id)

    def _generate_request_id(self, size: int = 3) -> int:
        """
        Generate random int of size in bytes
        :param size: Length in bytes
        :return:
        """
        while True:
            request_id = int.from_bytes(os.urandom(size), 'big')
            if request_id in list(self._async_response_map.keys()):
                continue
            return request_id

    def _async_response_resolver(self, request_id: int) -> Any:
        read_len = 0
        resolved = None
        while True:
            time.sleep(0.0001)
            responses = self._async_response_map[request_id][read_len:]
            read_len += len(responses)
            for item in responses:
                if item['response_type'] == TCResponseType.Error:
                    del self._async_response_map[request_id]
                    raise TonException(item['response_data'])
                if item['response_type'] == TCResponseType.Success:
                    resolved = item['response_data']
                if item['finished']:
                    del self._async_response_map[request_id]
                    return resolved

    async def _async_response_resolver_coro(self, request_id: int) -> Any:
        read_len = 0
        resolved = None
        while True:
            await asyncio.sleep(0.0001)
            responses = self._async_response_map[request_id][read_len:]
            read_len += len(responses)
            for item in responses:
                if item['response_type'] == TCResponseType.Error:
                    del self._async_response_map[request_id]
                    raise TonException(item['response_data'])
                if item['response_type'] == TCResponseType.Success:
                    resolved = item['response_data']
                if item['finished']:
                    del self._async_response_map[request_id]
                    return resolved

    def _async_response_generator(self, request_id: int) -> Any:
        read_len = 0
        while True:
            time.sleep(0.0001)
            responses = self._async_response_map[request_id][read_len:]
            read_len += len(responses)
            try:
                for item in responses:
                    if item['response_type'] == TCResponseType.Error:
                        del self._async_response_map[request_id]
                        raise TonException(item['response_data'])
                    if item['finished']:
                        del self._async_response_map[request_id]
                        if item['response_data']:
                            yield item
                        return
                    yield item
            except GeneratorExit:
                del self._async_response_map[request_id]
                return

    async def _async_response_generator_coro(self, request_id: int) -> Any:
        read_len = 0
        while True:
            await asyncio.sleep(0.0001)
            responses = self._async_response_map[request_id][read_len:]
            read_len += len(responses)
            try:
                for item in responses:
                    if item['response_type'] == TCResponseType.Error:
                        del self._async_response_map[request_id]
                        raise TonException(item['response_data'])
                    if item['finished']:
                        del self._async_response_map[request_id]
                        if item['response_data']:
                            yield item
                        return
                    yield item
            except GeneratorExit:
                del self._async_response_map[request_id]
                return

    @staticmethod
    def _prepare_params(params_or_str, **kwargs) -> str:
        """ Prepare params to pass to request """
        if isinstance(params_or_str, dict):
            params_or_str = {**params_or_str, **kwargs}
        elif params_or_str is None:
            params_or_str = kwargs or {}

        return json.dumps(params_or_str)

    @staticmethod
    @TCResponseHandler
    def _async_response_handler(
            request_id: int, response_data: TCStringData,
            response_type: int, finished: bool):
        """ Core response handler """
        response = {
            'response_data': response_data.json,
            'response_type': response_type,
            'finished': finished
        }
        logging.debug(f'Request: {request_id}; Response: {response}')
        if TonModule._async_response_map.get(request_id) is not None:
            TonModule._async_response_map[request_id].append(response)



from typing import Any


class TonException(Exception):
    def __init__(self, error: Any):
        if type(error) is dict:
            error = f"[{error.get('code')}] {error.get('message')} " \
                    f"(Core: {error.get('data', {}).get('core_version')})"
        super(TonException, self).__init__(error)



from typing import Any, Dict, Generator

from tonclient.decorators import Response
from tonclient.module import TonModule


class TonQLQuery(object):
    def __init__(self, collection: str):
        self._collection = collection
        self._filter = {}
        self._result = []
        self._order = None
        self._limit = None

    @property
    def collection(self) -> str:
        return self._collection

    @property
    def filter(self) -> Dict:
        return self._filter

    @property
    def result(self):
        return ' '.join(self._result)

    @property
    def order(self):
        return self._order

    @property
    def limit(self):
        return self._limit

    def set_filter(self, **kwargs: Any) -> 'TonQLQuery':
        """
        Generate GraphQL filter from kwargs.
        :param kwargs: kwarg name should be like string with field name
                    and condition, separated by "__".
                    kwarg value may be any accessible your want to filter by.
                    Full kwarg example: id__eq="0:000000"
        :return:
        """
        # Parse kwargs and generate GraphQL filter
        for field, value in kwargs.items():
            try:
                field, condition = field.split('__')
            except ValueError:
                raise ValueError(f"Incorrect '{field}' field syntax")

            if not self._filter.get(field):
                self._filter[field] = {}

            self._filter[field].update({condition: value})

        return self

    def set_result(self, fields: str = None, *args: str) -> 'TonQLQuery':
        """
        :param fields: Field names as a string splitted by space
        :param args: Any amount of table field names you want to include
                    in response
        :return:
        """
        self._result = [fields, *args]
        return self

    def set_order(self, fields: str, *args: str) -> 'TonQLQuery':
        """
        :param fields: Field names as a string splitted by space
                    field name with "-" to make DESC.
                    Example: "id" -> ASC, "-id" -> DESC
        :param args: Filed names as separate args
        :return:
        """
        fields = [*fields.split(' '), *args]
        if fields and not self._order:
            self._order = []

        for field in fields:
            direction = 'ASC'
            if field[0] == '-':
                direction = 'DESC'
                field = field[1:]
            self._order.append({'path': field, 'direction': direction})
        return self

    def set_limit(self, limit: int) -> 'TonQLQuery':
        self._limit = limit
        return self


class TonNet(TonModule):
    """ Free TON net SDK API implementation """
    @Response.query_collection
    def query_collection(self, query: TonQLQuery) -> Any:
        """ Queries collection data """
        return self.request(
            method='net.query_collection', collection=query.collection,
            filter=query.filter, result=query.result, order=query.order,
            limit=query.limit)

    @Response.wait_for_collection
    def wait_for_collection(
            self, query: TonQLQuery, timeout: int = None) -> Any:
        """
        Returns an object that fulfills the conditions or waits for
        its appearance
        """
        return self.request(
            method='net.wait_for_collection',
            collection=query.collection, filter=query.filter,
            result=query.result, timeout=timeout)

    @Response.subscribe_collection
    def subscribe_collection(self, query: TonQLQuery) -> Generator:
        """ Creates a subscription """
        return self.request(
            method='net.subscribe_collection', as_iterable=True,
            collection=query.collection, filter=query.filter,
            result=query.result)

    def unsubscribe(self, handle: int):
        """ Cancels a subscription """
        return self.request(method='net.unsubscribe', handle=handle)



import unittest

from tonclient.bindings.lib import LIB_VERSION
from tonclient.client import TonClient, DEVNET_BASE_URL


class TestTonClientAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        self.client = TonClient(network={'server_address': DEVNET_BASE_URL})

    def test_version(self):
        self.assertEqual(LIB_VERSION, self.client.version())

    def test_get_api_reference(self):
        reference = self.client.get_api_reference()
        self.assertEqual(LIB_VERSION, reference['version'])

    def test_build_info(self):
        info = self.client.build_info()
        self.assertIsInstance(info, dict)

    def test_destroy_context(self):
        self.client.destroy_context()


class TestTonClientSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use """
    def setUp(self) -> None:
        self.client = TonClient(
            network={'server_address': DEVNET_BASE_URL}, is_core_async=False)

    def test_version(self):
        self.assertEqual(LIB_VERSION, self.client.version())

    def test_get_api_reference(self):
        reference = self.client.get_api_reference()
        self.assertEqual(LIB_VERSION, reference['version'])

    def test_build_info(self):
        info = self.client.build_info()
        self.assertIsInstance(info, dict)

    def test_destroy_context(self):
        self.client.destroy_context()



from typing import Dict, Any

from tonclient.decorators import Response
from tonclient.module import TonModule


class TonBoc(TonModule):
    """ Free TON boc SDK API implementation """
    @Response.parse_message
    def parse_message(self, boc: str) -> Dict[str, Any]:
        """
        Parses message BOC into a JSON
        :param boc: Base64 encoded BOC
        :return:
        """
        return self.request(method='boc.parse_message', boc=boc)

    @Response.parse_transaction
    def parse_transaction(self, boc: str) -> Dict[str, Any]:
        """
        Parses transaction BOC into a JSON
        :param boc: Base64 encoded BOC
        :return:
        """
        return self.request(method='boc.parse_transaction', boc=boc)

    @Response.parse_account
    def parse_account(self, boc: str) -> Dict[str, Any]:
        """
        Parses account BOC into a JSON
        :param boc: Base64 encoded BOC
        :return:
        """
        return self.request(method='boc.parse_account', boc=boc)

    @Response.parse_block
    def parse_block(self, boc: str) -> Dict[str, Any]:
        """
        Parses block BOC into a JSON
        :param boc: Base64 encoded BOC
        :return:
        """
        return self.request(method='boc.parse_block', boc=boc)

    @Response.get_blockchain_config
    def get_blockchain_config(self, block_boc: str) -> str:
        """
        :param block_boc: Base64 encoded block boc
        :return:
        """
        return self.request(
            method='boc.get_blockchain_config', block_boc=block_boc)



import base64
import os
import unittest

from tonclient.errors import TonException
from tonclient.test.helpers import SAMPLES_DIR, async_core_client, \
    sync_core_client
from tonclient.types import Abi, KeyPair, DeploySet, CallSet, Signer, \
    MessageSource, StateInitSource


class TestTonAbiAsyncCore(unittest.TestCase):
    def setUp(self) -> None:
        # Events contract params
        self.events_abi = Abi.from_json_path(
            path=os.path.join(SAMPLES_DIR, 'Events.abi.json'))
        self.keypair = KeyPair(
            public='4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499',
            secret='cc8929d635719612a9478b9cd17675a39cfad52d8959e8a177389b8c0b9122a7')
        with open(os.path.join(SAMPLES_DIR, 'Events.tvc'), 'rb') as fp:
            self.events_tvc = base64.b64encode(fp.read()).decode()
        self.events_time = 1599458364291
        self.events_expire = 1599458404

    def test_decode_message(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        decoded = async_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Input', 'name': 'returnValue', 'value': {'id': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': {'expire': 1599458404, 'time': 1599458364291, 'pubkey': '4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499'}},
            decoded)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMJL6z6ro48sYvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        decoded = async_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Event', 'name': 'EventThrown', 'value': {'id': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': None},
            decoded)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMKr6z6rxK3xYJAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        decoded = async_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Output', 'name': 'returnValue', 'value': {'value0': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': None},
            decoded)

        with self.assertRaises(TonException):
            async_core_client.abi.decode_message(
                abi=self.events_abi, message='0x0')

    def test_decode_message_body(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        parsed = async_core_client.boc.parse_message(boc=message)
        decoded = async_core_client.abi.decode_message_body(
            abi=self.events_abi, body=parsed['body'])
        self.assertEqual(
            {'body_type': 'Input', 'name': 'returnValue', 'value': {'id': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': {'expire': 1599458404, 'time': 1599458364291, 'pubkey': '4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499'}},
            decoded)

    def test_encode_message(self):
        deploy_set = DeploySet(tvc=self.events_tvc)
        call_set = CallSet(
            function_name='constructor',
            header={
                'pubkey': self.keypair.public,
                'time': self.events_time,
                'expire': self.events_expire
            })
        signer = Signer.from_external(public=self.keypair.public)

        # Create unsigned deployment message
        unsigned = async_core_client.abi.encode_message(
            abi=self.events_abi, deploy_set=deploy_set, call_set=call_set,
            signer=signer)
        self.assertEqual(
            'te6ccgECFwEAA2gAAqeIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEZTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gGAQEBwAICA88gBQMBAd4EAAPQIABB2mPiBH+O713GsgL3S844tQp+62YECSCD0w6eEqy4TKTMAib/APSkICLAAZL0oOGK7VNYMPShCQcBCvSkIPShCAAAAgEgDAoByP9/Ie1E0CDXScIBjhDT/9M/0wDRf/hh+Gb4Y/hijhj0BXABgED0DvK91wv/+GJw+GNw+GZ/+GHi0wABjh2BAgDXGCD5AQHTAAGU0/8DAZMC+ELiIPhl+RDyqJXTAAHyeuLTPwELAGqOHvhDIbkgnzAg+COBA+iogggbd0Cgud6S+GPggDTyNNjTHwH4I7zyudMfAfAB+EdukvI83gIBIBINAgEgDw4AvbqLVfP/hBbo417UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLe+Ebyc3H4ZtH4APhCyMv/+EPPCz/4Rs8LAMntVH/4Z4AgEgERAA5biABrW/CC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb30gyupo6H0gb+j8IpA3SRg4b3whXXlwMnwAZGT9ghBkZ8KEZ0aCBAfQAAAAAAAAAAAAAAAAACBni2TAgEB9gBh8IWRl//wh54Wf/CNnhYBk9qo//DPAAxbmTwqLfCC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb2uG/8rqaOhp/+/o/ABkRe4AAAAAAAAAAAAAAAAIZ4tnwOfI48sYvRDnhf/kuP2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8AIBSBYTAQm4t8WCUBQB/PhBbo4T7UTQ0//TP9MA0X/4Yfhm+GP4Yt7XDf+V1NHQ0//f0fgAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPkceWMXohzwv/yXH7AMiL3AAAAAAAAAAAAAAAABDPFs+Bz5JW+LBKIc8L/8lx+wAw+ELIy//4Q88LP/hGzwsAye1UfxUABPhnAHLccCLQ1gIx0gAw3CHHAJLyO+Ah1w0fkvI84VMRkvI74cEEIoIQ/////byxkvI84AHwAfhHbpLyPN4=',
            unsigned['message'])
        self.assertEqual(
            'KCGM36iTYuCYynk+Jnemis+mcwi3RFCke95i7l96s4Q=',
            unsigned['data_to_sign'])

        # Create detached signature
        signature = async_core_client.crypto.sign(
            unsigned=unsigned['data_to_sign'], keys=self.keypair)
        self.assertEqual(
            '6272357bccb601db2b821cb0f5f564ab519212d242cf31961fe9a3c50a30b236012618296b4f769355c0e9567cd25b366f3c037435c498c82e5305622adbc70e',
            signature['signature'])

        # Attach signature to unsigned message
        signed = async_core_client.abi.attach_signature(
            abi=self.events_abi, public_key=self.keypair.public,
            message=unsigned['message'], signature=signature['signature'])
        self.assertEqual(
            'te6ccgECGAEAA6wAA0eIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEbAHAgEA4bE5Gr3mWwDtlcEOWHr6slWoyQlpIWeYyw/00eKFGFkbAJMMFLWnu0mq4HSrPmktmzeeAboa4kxkFymCsRVt44dTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gAQHAAwIDzyAGBAEB3gUAA9AgAEHaY+IEf47vXcayAvdLzji1Cn7rZgQJIIPTDp4SrLhMpMwCJv8A9KQgIsABkvSg4YrtU1gw9KEKCAEK9KQg9KEJAAACASANCwHI/38h7UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLTAAGOHYECANcYIPkBAdMAAZTT/wMBkwL4QuIg+GX5EPKoldMAAfJ64tM/AQwAao4e+EMhuSCfMCD4I4ED6KiCCBt3QKC53pL4Y+CANPI02NMfAfgjvPK50x8B8AH4R26S8jzeAgEgEw4CASAQDwC9uotV8/+EFujjXtRNAg10nCAY4Q0//TP9MA0X/4Yfhm+GP4Yo4Y9AVwAYBA9A7yvdcL//hicPhjcPhmf/hh4t74RvJzcfhm0fgA+ELIy//4Q88LP/hGzwsAye1Uf/hngCASASEQDluIAGtb8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFvfSDK6mjofSBv6PwikDdJGDhvfCFdeXAyfABkZP2CEGRnwoRnRoIEB9AAAAAAAAAAAAAAAAAAIGeLZMCAQH2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8ADFuZPCot8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFva4b/yupo6Gn/7+j8AGRF7gAAAAAAAAAAAAAAAAhni2fA58jjyxi9EOeF/+S4/YAYfCFkZf/8IeeFn/wjZ4WAZPaqP/wzwAgFIFxQBCbi3xYJQFQH8+EFujhPtRNDT/9M/0wDRf/hh+Gb4Y/hi3tcN/5XU0dDT/9/R+ADIi9wAAAAAAAAAAAAAAAAQzxbPgc+Rx5YxeiHPC//JcfsAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPklb4sEohzwv/yXH7ADD4QsjL//hDzws/+EbPCwDJ7VR/FgAE+GcActxwItDWAjHSADDcIccAkvI74CHXDR+S8jzhUxGS8jvhwQQighD////9vLGS8jzgAfAB+EdukvI83g==',
            signed['message'])

        # Create initially signed message
        signer = Signer.from_keypair(keypair=self.keypair)
        signed = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        self.assertEqual(
            'te6ccgECGAEAA6wAA0eIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEbAHAgEA4bE5Gr3mWwDtlcEOWHr6slWoyQlpIWeYyw/00eKFGFkbAJMMFLWnu0mq4HSrPmktmzeeAboa4kxkFymCsRVt44dTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gAQHAAwIDzyAGBAEB3gUAA9AgAEHaY+IEf47vXcayAvdLzji1Cn7rZgQJIIPTDp4SrLhMpMwCJv8A9KQgIsABkvSg4YrtU1gw9KEKCAEK9KQg9KEJAAACASANCwHI/38h7UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLTAAGOHYECANcYIPkBAdMAAZTT/wMBkwL4QuIg+GX5EPKoldMAAfJ64tM/AQwAao4e+EMhuSCfMCD4I4ED6KiCCBt3QKC53pL4Y+CANPI02NMfAfgjvPK50x8B8AH4R26S8jzeAgEgEw4CASAQDwC9uotV8/+EFujjXtRNAg10nCAY4Q0//TP9MA0X/4Yfhm+GP4Yo4Y9AVwAYBA9A7yvdcL//hicPhjcPhmf/hh4t74RvJzcfhm0fgA+ELIy//4Q88LP/hGzwsAye1Uf/hngCASASEQDluIAGtb8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFvfSDK6mjofSBv6PwikDdJGDhvfCFdeXAyfABkZP2CEGRnwoRnRoIEB9AAAAAAAAAAAAAAAAAAIGeLZMCAQH2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8ADFuZPCot8ILdHCfaiaGn/6Z/pgGi//DD8M3wx/DFva4b/yupo6Gn/7+j8AGRF7gAAAAAAAAAAAAAAAAhni2fA58jjyxi9EOeF/+S4/YAYfCFkZf/8IeeFn/wjZ4WAZPaqP/wzwAgFIFxQBCbi3xYJQFQH8+EFujhPtRNDT/9M/0wDRf/hh+Gb4Y/hi3tcN/5XU0dDT/9/R+ADIi9wAAAAAAAAAAAAAAAAQzxbPgc+Rx5YxeiHPC//JcfsAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPklb4sEohzwv/yXH7ADD4QsjL//hDzws/+EbPCwDJ7VR/FgAE+GcActxwItDWAjHSADDcIccAkvI74CHXDR+S8jzhUxGS8jvhwQQighD////9vLGS8jzgAfAB+EdukvI83g==',
            signed['message'])

        # Create run unsigned message
        address = '0:05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309'
        call_set = CallSet(
            function_name='returnValue', inputs={'id': '0'},
            header={
                'pubkey': self.keypair.public,
                'time': self.events_time,
                'expire': self.events_expire
            })
        signer = Signer.from_external(public=self.keypair.public)
        unsigned = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer, address=address,
            call_set=call_set)
        self.assertEqual(
            'te6ccgEBAgEAeAABpYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIFMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKAQBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=',
            unsigned['message'])
        self.assertEqual(
            'i4Hs3PB12QA9UBFbOIpkG3JerHHqjm4LgvF4MA7TDsY=',
            unsigned['data_to_sign'])

        # Create detached signature
        signature = async_core_client.crypto.sign(
            unsigned=unsigned['data_to_sign'], keys=self.keypair)
        self.assertEqual(
            '5bbfb7f184f2cb5f019400b9cd497eeaa41f3d5885619e9f7d4fab8dd695f4b3a02159a1422996c1dd7d1be67898bc79c6adba6c65a18101ac5f0a2a2bb8910b',
            signature['signature'])

        # Attach signature
        signed = async_core_client.abi.attach_signature(
            abi=self.events_abi, public_key=self.keypair.public,
            message=unsigned['message'], signature=signature['signature'])
        self.assertEqual(
            'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==',
            signed['message'])

        # Create initially signed message
        signer = Signer.from_keypair(keypair=self.keypair)
        signed = async_core_client.abi.encode_message(
            abi=self.events_abi, signer=signer, address=address,
            call_set=call_set)
        self.assertEqual(
            'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==',
            signed['message'])

    def test_encode_account(self):
        # Encode account from encoded deploy message
        encoded_deploy_message = 'te6ccgECFwEAA2gAAqeIAAt9aqvShfTon7Lei1PVOhUEkEEZQkhDKPgNyzeTL6YSEZTHxAj/Hd67jWQF7peccWoU/dbMCBJBB6YdPCVZcJlJkAAAF0ZyXLg19VzGRotV8/gGAQEBwAICA88gBQMBAd4EAAPQIABB2mPiBH+O713GsgL3S844tQp+62YECSCD0w6eEqy4TKTMAib/APSkICLAAZL0oOGK7VNYMPShCQcBCvSkIPShCAAAAgEgDAoByP9/Ie1E0CDXScIBjhDT/9M/0wDRf/hh+Gb4Y/hijhj0BXABgED0DvK91wv/+GJw+GNw+GZ/+GHi0wABjh2BAgDXGCD5AQHTAAGU0/8DAZMC+ELiIPhl+RDyqJXTAAHyeuLTPwELAGqOHvhDIbkgnzAg+COBA+iogggbd0Cgud6S+GPggDTyNNjTHwH4I7zyudMfAfAB+EdukvI83gIBIBINAgEgDw4AvbqLVfP/hBbo417UTQINdJwgGOENP/0z/TANF/+GH4Zvhj+GKOGPQFcAGAQPQO8r3XC//4YnD4Y3D4Zn/4YeLe+Ebyc3H4ZtH4APhCyMv/+EPPCz/4Rs8LAMntVH/4Z4AgEgERAA5biABrW/CC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb30gyupo6H0gb+j8IpA3SRg4b3whXXlwMnwAZGT9ghBkZ8KEZ0aCBAfQAAAAAAAAAAAAAAAAACBni2TAgEB9gBh8IWRl//wh54Wf/CNnhYBk9qo//DPAAxbmTwqLfCC3Rwn2omhp/+mf6YBov/ww/DN8Mfwxb2uG/8rqaOhp/+/o/ABkRe4AAAAAAAAAAAAAAAAIZ4tnwOfI48sYvRDnhf/kuP2AGHwhZGX//CHnhZ/8I2eFgGT2qj/8M8AIBSBYTAQm4t8WCUBQB/PhBbo4T7UTQ0//TP9MA0X/4Yfhm+GP4Yt7XDf+V1NHQ0//f0fgAyIvcAAAAAAAAAAAAAAAAEM8Wz4HPkceWMXohzwv/yXH7AMiL3AAAAAAAAAAAAAAAABDPFs+Bz5JW+LBKIc8L/8lx+wAw+ELIy//4Q88LP/hGzwsAye1UfxUABPhnAHLccCLQ1gIx0gAw3CHHAJLyO+Ah1w0fkvI84VMRkvI74cEEIoIQ/////byxkvI84AHwAfhHbpLyPN4='
        message_source = MessageSource.from_encoded(
            message=encoded_deploy_message, abi=self.events_abi)
        state_init_source = StateInitSource.from_message(
            message=message_source)

        encoded = async_core_client.abi.encode_account(
            state_init=state_init_source)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded['id'])

        # Encode account from encoding params
        deploy_set = DeploySet(tvc=self.events_tvc)
        call_set = CallSet(
            function_name='constructor',
            header={
                'pubkey': self.keypair.public,
                'time': self.events_time,
                'expire': self.events_expire
            })
        signer = Signer.from_keypair(keypair=self.keypair)
        message_source = MessageSource.from_encoding_params(
            abi=self.events_abi, signer=signer, deploy_set=deploy_set,
            call_set=call_set)
        state_init_source = StateInitSource.from_message(
            message=message_source)
        encoded = async_core_client.abi.encode_account(
            state_init=state_init_source)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded['id'])

        # Test exception (external signer)
        with self.assertRaises(TonException):
            signer = Signer.from_external(public=self.keypair.public)
            message_source = MessageSource.from_encoding_params(
                abi=self.events_abi, signer=signer, deploy_set=deploy_set,
                call_set=call_set)
            state_init_source = StateInitSource.from_message(
                message=message_source)
            async_core_client.abi.encode_account(state_init=state_init_source)

        # Encode account from TVC
        state_init_source = StateInitSource.from_tvc(tvc=self.events_tvc)
        encoded = async_core_client.abi.encode_account(
            state_init=state_init_source)
        self.assertNotEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded['id'])

        state_init_source = StateInitSource.from_tvc(
            tvc=self.events_tvc, public_key=self.keypair.public)
        encoded = async_core_client.abi.encode_account(
            state_init=state_init_source)
        self.assertEqual(
            '05beb555e942fa744fd96f45a9ea9d0a8248208ca12421947c06e59bc997d309',
            encoded['id'])


class TestTonAbiSyncCore(unittest.TestCase):
    """ Sync core is not recommended to use, so make just a couple of tests """
    def setUp(self) -> None:
        # Events contract params
        self.events_abi = Abi.from_json_path(
            path=os.path.join(SAMPLES_DIR, 'Events.abi.json'))
        self.keypair = KeyPair(
            public='4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499',
            secret='cc8929d635719612a9478b9cd17675a39cfad52d8959e8a177389b8c0b9122a7')
        with open(os.path.join(SAMPLES_DIR, 'Events.tvc'), 'rb') as fp:
            self.events_tvc = base64.b64encode(fp.read()).decode()
        self.events_time = 1599458364291
        self.events_expire = 1599458404

    def test_decode_message(self):
        message = 'te6ccgEBAwEAvAABRYgAC31qq9KF9Oifst6LU9U6FQSQQRlCSEMo+A3LN5MvphIMAQHhrd/b+MJ5Za+AygBc5qS/dVIPnqxCsM9PvqfVxutK+lnQEKzQoRTLYO6+jfM8TF4841bdNjLQwIDWL4UVFdxIhdMfECP8d3ruNZAXul5xxahT91swIEkEHph08JVlwmUmQAAAXRnJcuDX1XMZBW+LBKACAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=='
        decoded = sync_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Input', 'name': 'returnValue', 'value': {'id': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': {'expire': 1599458404, 'time': 1599458364291, 'pubkey': '4c7c408ff1ddebb8d6405ee979c716a14fdd6cc08124107a61d3c25597099499'}},
            decoded)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMJL6z6ro48sYvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        decoded = sync_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Event', 'name': 'EventThrown', 'value': {'id': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': None},
            decoded)

        message = 'te6ccgEBAQEAVQAApeACvg5/pmQpY4m61HmJ0ne+zjHJu3MNG8rJxUDLbHKBu/AAAAAAAAAMKr6z6rxK3xYJAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABA'
        decoded = sync_core_client.abi.decode_message(
            abi=self.events_abi, message=message)
        self.assertEqual(
            {'body_type': 'Output', 'name': 'returnValue', 'value': {'value0': '0x0000000000000000000000000000000000000000000000000000000000000000'}, 'header': None},
            decoded)

        with self.assertRaises(TonException):
            sync_core_client.abi.decode_message(
                abi=self.events_abi, message='0x0')



import json
from io import StringIO
from typing import Dict, Union, Any

DEFAULT_MNEMONIC_DICTIONARY = 1
DEFAULT_MNEMONIC_WORD_COUNT = 12
DEFAULT_HDKEY_DERIVATION_PATH = "m/44'/396'/0'/0/0"


class MnemonicDictionary(object):
    TON = 0
    ENGLISH = 1
    CHINESE_SIMPLIFIED = 2
    CHINESE_TRADITIONAL = 3
    FRENCH = 4
    ITALIAN = 5
    JAPANESE = 6
    KOREAN = 7
    SPANISH = 8


class KeyPair(object):
    """ Keypair object representation """
    def __init__(self, public: str, secret: str):
        self.public = public
        self.secret = secret

    @property
    def dict(self) -> Dict[str, str]:
        return {'public': self.public, 'secret': self.secret}

    @property
    def binary(self) -> bytes:
        return bytes.fromhex(f'{self.secret}{self.public}')

    @staticmethod
    def load(path: str, is_binary: bool) -> 'KeyPair':
        """ Load keypair from file """
        if is_binary:
            with open(path, 'rb') as fp:
                keys = fp.read().hex()
                keys = {'public': keys[64:], 'secret': keys[:64]}
        else:
            with open(path, 'r') as fp:
                keys = json.loads(fp.read())

        return KeyPair(**keys)

    def dump(self, path: str, as_binary: bool) -> None:
        """ Dump keypair to file """
        if as_binary:
            with open(path, 'wb') as fp:
                fp.write(self.binary)
        else:
            with open(path, 'w') as fp:
                json.dump(self.dict, fp)

    @staticmethod
    def load_io(io: StringIO, as_binary: bool = False) -> 'KeyPair':
        """ Load keypair from StringIO """
        data = io.getvalue()
        keys = json.loads(data) \
            if not as_binary else {'public': data[64:], 'secret': data[:64]}

        return KeyPair(**keys)

    def dump_io(self, io: StringIO, as_binary: bool = False):
        """ Dump keypair to StringIO """
        keys = json.dumps(self.dict) if not as_binary else self.binary.hex()
        io.write(keys)


class Abi(object):
    def __init__(self, abi: Union[Dict, str, int], abi_type: str):
        """
        :param abi: ABI dict, ABI JSON string or handle
        :param abi_type: One of 'Serialized', 'Handle'
        """
        self.abi = abi
        self.abi_type = abi_type

    @property
    def dict(self):
        if type(self.abi) is str:
            self.abi = json.loads(self.abi)
        return {'type': self.abi_type, 'value': self.abi}

    @staticmethod
    def from_dict(abi: Dict[str, Any]) -> 'Abi':
        return Abi(abi=abi, abi_type='Serialized')

    @staticmethod
    def from_json_string(abi_json: str) -> 'Abi':
        return Abi(abi=json.loads(abi_json), abi_type='Serialized')

    @staticmethod
    def from_json_path(path: str) -> 'Abi':
        with open(path, 'r') as fp:
            return Abi(abi=json.loads(fp.read()), abi_type='Serialized')

    @staticmethod
    def from_handle(handle: int) -> 'Abi':
        return Abi(abi=handle, abi_type='Handle')


class Signer(object):
    def __init__(
            self, keypair: KeyPair = None, public: str = None,
            box_handle: int = None):
        """
        :param keypair: message will be signed using the provided keys
        :param public: message will be signed using external methods.
                Public key must be provided with 'hex' encoding
        :param box_handle: message will be signed using the provided
                signing box
        """
        self.keypair = keypair
        self.public = public
        self.box_handle = box_handle

    @property
    def dict(self):
        if self.keypair:
            return {'type': 'Keys', 'keys': self.keypair.dict}
        elif self.public:
            return {'type': 'External', 'public_key': self.public}
        elif self.box_handle:
            return {'type': 'SigningBox', 'handle': self.box_handle}
        return {'type': 'None'}

    @staticmethod
    def none() -> 'Signer':
        return Signer()

    @staticmethod
    def from_keypair(keypair: KeyPair) -> 'Signer':
        return Signer(keypair=keypair)

    @staticmethod
    def from_external(public: str) -> 'Signer':
        return Signer(public=public)

    @staticmethod
    def from_signing_box(box_handle: int) -> 'Signer':
        return Signer(box_handle=box_handle)


class DeploySet(object):
    def __init__(
            self, tvc: str, workchain_id: int = 0,
            initial_data: Dict[str, Any] = None):
        """
        :param tvc: Base64 encoded TVC file content
        :param workchain_id: Target workchain for destination address
        :param initial_data: Initial values for contract's public variables
        """
        self.tvc = tvc
        self.workchain_id = workchain_id
        self.initial_data = initial_data

    @property
    def dict(self):
        return {
            'tvc': self.tvc,
            'workchain_id': self.workchain_id,
            'initial_data': self.initial_data
        }


class CallSet(object):
    def __init__(
            self, function_name: str, header: Dict[str, Any] = None,
            inputs: Dict[str, Any] = None):
        """
        :param function_name: Function name
        :param header: Function header. If an application omit some parameters
                required by the contract's ABI, the library will set the
                default values for it
        :param inputs: Function input according to ABI
        """
        self.function_name = function_name
        self.header = header
        self.input = inputs

    @property
    def dict(self):
        return {
            'function_name': self.function_name,
            'header': self.header,
            'input': self.input
        }


class MessageSource(object):
    def __init__(self, source_type: str, **kwargs):
        """
        :param source_type: One of 'Encoded', 'EncodingParams'
        :param kwargs:
        """
        self.source_type = source_type
        self._kwargs = kwargs
        for name, value in kwargs.items():
            setattr(self, name, value)

    @property
    def dict(self):
        params = {
            key: value.dict if hasattr(value, 'dict') else value
            for key, value in self._kwargs.items()
        }
        return {'type': self.source_type, **params}

    @staticmethod
    def from_encoded(message: str, abi: Abi = None) -> 'MessageSource':
        return MessageSource(source_type='Encoded', message=message, abi=abi)

    @staticmethod
    def from_encoding_params(**kwargs) -> 'MessageSource':
        """
        :param kwargs: The same kwargs as for 'abi.encode_message' method
        :return:
        """
        return MessageSource(source_type='EncodingParams', **kwargs)


class StateInitSource(object):
    def __init__(self, source_type: str, **kwargs):
        """
        :param source_type: One of 'Message', 'StateInit', 'Tvc'
        """
        self.source_type = source_type
        self._kwargs = kwargs

    @property
    def dict(self):
        return {'type': self.source_type, **self._kwargs}

    @staticmethod
    def from_message(message: MessageSource) -> 'StateInitSource':
        """
        :param message: Deploy message
        :return:
        """
        return StateInitSource(source_type='Message', source=message.dict)

    @staticmethod
    def from_state_init(
            code: str, data: str, library: str = None) -> 'StateInitSource':
        """
        :param code: Base64 encoded code BOC
        :param data: Base64 encoded data BOC
        :param library: Base64 encoded library BOC
        :return:
        """
        return StateInitSource(
            source_type='StateInit', code=code, data=data, library=library)

    @staticmethod
    def from_tvc(
            tvc: str, public_key: str = None, abi: Abi = None,
            value: Any = None) -> 'StateInitSource':
        """
        :param tvc: Base64 encoded TVC data
        :param public_key: Hex encoded public key
        :param abi: Contract ABI
        :param value: Value to pass. If provided 'abi' is required
        :return:
        """
        init_params = None
        if abi and value:
            init_params = {'abi': abi.dict, 'value': value}
        return StateInitSource(
            source_type='Tvc', tvc=tvc, public_key=public_key,
            init_params=init_params)


class AddressStringFormat(object):
    AccountId = {'type': 'AccountId'}
    Hex = {'type': 'Hex'}

    @staticmethod
    def base64(
            url: bool = False, test: bool = False, bounce: bool = False
    ) -> Dict[str, Union[str, bool]]:
        return {'type': 'Base64', 'url': url, 'test': test, 'bounce': bounce}


class AccountForExecutor(object):
    def __init__(
            self, uninit: bool = False, boc: str = None,
            unlimited_balance: bool = False):
        self._uninit = uninit
        self._boc = boc
        self._unlimited_balance = unlimited_balance

    @property
    def dict(self) -> Dict[str, Any]:
        if self._uninit:
            return {'type': 'Uninit'}
        elif self._boc:
            return {
                'type': 'Account',
                'boc': self._boc,
                'unlimited_balance': self._unlimited_balance
            }
        return {'type': 'None'}

    @staticmethod
    def none() -> 'AccountForExecutor':
        return AccountForExecutor()

    @staticmethod
    def uninit() -> 'AccountForExecutor':
        """
        Emulate uninitialized account to run deploy message
        :return:
        """
        return AccountForExecutor(uninit=True)

    @staticmethod
    def from_account(
            boc: str, unlimited_balance: bool = False
    ) -> 'AccountForExecutor':
        """
        :param boc: Base64 encoded account BOC
        :param unlimited_balance: Flag for running account with the unlimited
                balance. Can be used to calculate transaction fees without
                balance check
        :return:
        """
        return AccountForExecutor(boc=boc, unlimited_balance=unlimited_balance)
